//******************************************************************************
// ED Text Editor
// Version 0.12
// Copyright (c) 2011, Dave Hein
// See end of file for terms of use.
//******************************************************************************
#include <stdio.h>
#include <memory.h>
#include <string.h>
#include <stdlib.h>
#include "smalloc.h"

#define UNSPECIFIED   -2
#define PARSING_ERROR -3
#define TAB_CHAR 9

#ifdef CSPIN_FLAG
#define MEMORY_SIZE   16000
#else
#define MEMORY_SIZE   40000

extern StringT *smfreelist;
#endif

int FindString(char *str1, char *str2);

int numline = 0;
char srchstr[100];
char rplcstr1[100];
char rplcstr2[100];
char inbuf[200];
int scroll = 20;
int rflag = 0;
int gflag = 0;
int vmode = 0;
int vlines = 21;
int prompt = 0;
int changed = 0;
char *bigbuf;
StringT *ListHead;
StringT *YankHead;

// Allocate a string link and copy contents of "str"
StringT *AllocateString(char *str)
{
    StringT *link = smalloc(strlen(str) + 1);
    if (link) strcpy(link->str, str);
    return link;
}

// Duplicate a string link
StringT *DuplicateString(StringT *link)
{
    return AllocateString(link->str);
}

// Check for a memory leak
void CheckMemory(void)
{
    int total;
    StringT *curr;
    int listsize = 0;
    int yanksize = 0;
    int freesize = 0;

    for (curr = ListHead; curr; curr = curr->next)
    {
	listsize += curr->size;
    }

    for (curr = YankHead; curr; curr = curr->next)
    {
	yanksize += curr->size;
    }

    for (curr = smfreelist; curr; curr = curr->next)
    {
	freesize += curr->size;
    }

    total = listsize + yanksize + freesize;

    if (total != MEMORY_SIZE)
    {
	printf("Memory leak!\n");
        printf("listsize = %d, yanksize = %d, freesize = %d, total = %d\n",
            listsize, yanksize, freesize, listsize+yanksize+freesize);
    }
}

// Get a pointer to a string link from its line number
StringT *GetLinePtr(int linenum, StringP *pprev)
{
    StringT *prev = 0;
    StringT *curr = ListHead;

    while (curr && --linenum > 0)
    {
	prev = curr;
	curr = curr->next;
    }

    *pprev = prev;
    return curr;
}

// Remove the CR and LF characters from the end of a line
void RemoveCRLF(char *str)
{
    int len = strlen(str);

    str += len - 1;

    while (len > 0)
    {
	if (*str != 10 && *str != 13) break;
	*str-- = 0;
	len--;
    }
}

// Insert a string link at the line given by "linenum"
void InsertLine(StringT *line, int linenum)
{
    StringT *prev;
    StringT *curr = GetLinePtr(linenum-1, &prev);

    if (linenum < 1 || linenum > numline + 1)
        printf("InsertLine: linenum = %d, numline = %d\n", linenum, numline);

    if (curr == 0 && (linenum != 1 || ListHead != 0))
        printf("InsertLine: curr == 0, linenum = %d\n", linenum);

    if (linenum == 1)
    {
	line->next = ListHead;
	ListHead = line;
    }
    else
    {
	line->next = curr->next;
	curr->next = line;
    }

    numline++;
}

// Delete the line at "linenum"
void DeleteLine(int linenum)
{
    StringT *prev;
    StringT *curr = GetLinePtr(linenum, &prev);
    if (!curr)
    {
	printf("DeleteLine: curr = 0, linenum = %d\n", linenum);
	return;
    }
    if (!prev) ListHead   = curr->next;
    else       prev->next = curr->next;
    sfree(curr);
    numline--;
}

// Based on the value of gflag, check for if the line at "lineptr"
// does, or does not contains "str"
int CheckMatch(char *lineptr, char *str)
{
    if (gflag == 0)
        return 0;

    if (gflag == 1)
    {
	if (!FindString(lineptr, str)) return 1;
	return 0;
    }

    if (FindString(lineptr, str)) return 1;
    return 0;
}

// Print lines using the "n", "v", "l" or "p" format
int PrintLines(int linenum, int first, int last, int mode)
{
    int line;
    int fflag;
    int lastproc = linenum;
    char *LinePtr;
    StringT *prev;
    StringT *curr = GetLinePtr(first, &prev);

    for (line = first; line <= last; line++)
    {
        if (!curr)
        {
	    printf("PrintLines: curr = 0, line = %s\n", line);
	    return 0;
        }
	LinePtr = curr->str;
	curr = curr->next;
	if (CheckMatch(LinePtr, srchstr)) continue;
	if (mode == 'n') printf("%6d  %s\n", line, LinePtr);
	else if (mode == 'v')
	{
	    if (line == linenum)
	        printf("%6d >%s\n", line, LinePtr);
	    else
	        printf("%6d  %s\n", line, LinePtr);
	}
	else if (mode == 'l')
	{
	    while (*LinePtr)
	    {
		if (*LinePtr >= 32 && *LinePtr < 127)
		    printf("%c", *LinePtr);
		else if (*LinePtr == TAB_CHAR)
		    printf("\\t");
		else
		    printf("\\x%2.2x", *LinePtr);
		LinePtr++;
	    }
	    printf("$\n");
	}
	else printf("%s\n", LinePtr);
	lastproc = line;
    }
    return lastproc;
}

#ifndef CSPIN_FLAG
// Return true if val is an ASCII decimal digit
int isdigit(int val)
{
    return val >= '0' && val <= '9';
}
#endif

// Decode a number from the string
int getnum(char **pstr)
{
    int val = 0;
    char *str = *pstr;

    while (isdigit(*str))
    {
	val = (val * 10) + *str++ - '0';
    }

    *pstr = str;
    return val;
}

// Find the next character that matches "val"
char *FindChar(char *str, int val)
{
    while (*str)
    {
	if (*str == val) break;
	str++;
    }
    return str;
}

// Skip the next character that does not match "val"
char *SkipChar(char *str, int val)
{
    while (*str)
    {
	if (*str != val) break;
	str++;
    }
    return str;
}

// Return true if "str2" is found in "str1"
int FindString(char *str1, char *str2)
{
    int len;

    len = strlen(str2);

    while (*str1)
    {
	if (strncmp(str1, str2, len) == 0) return 1;
	str1++;
    }
    return 0;
}

// Use the first character as a delimeter and copy the string to srchstr
int GetString(char **pstr)
{
    char *ptr;
    char *str = *pstr;
    int val = *str++;
    int len;

    if (!val) return 0;

    ptr = FindChar(str, val);
    len = ((int)ptr) - ((int)str);
    if (*ptr) ptr++;
    *pstr = ptr;
    if (len > 0)
    {
        memcpy(srchstr, str, len);
        srchstr[len] = 0;
    }
    return 1;
}

// Search forward through all the lines for the first occueance of srchstr
int SearchStringForward(char **pstr, int linenum)
{
    int i;
    StringT *prev;
    StringT *curr = GetLinePtr(linenum + 1, &prev);

    if (!GetString(pstr)) return PARSING_ERROR;

    for (i = linenum + 1; i <= numline; i++)
    {
	if (!curr) { printf("SearchStringForward: curr = 0, line = %d\n", i); return -3;}
	if (FindString(curr->str, srchstr)) return i;
	curr = curr->next;
    }

    curr = ListHead;
    for (i = 1; i <= linenum; i++)
    {
	if (!curr) { printf("SearchStringForward: curr = 0, line = %d\n", i); return -3;}
	if (FindString(curr->str, srchstr)) return i;
	curr = curr->next;
    }

    return PARSING_ERROR;
}

// Search backward through all the lines for the first occueance of srchstr
int SearchStringBackward(char **pstr, int linenum)
{
    int i;
    StringT *prev;
    StringT *curr;

    if (!GetString(pstr)) return PARSING_ERROR;

    for (i = linenum - 1; i >= 1; i--)
    {
        curr = GetLinePtr(i, &prev);
	if (!curr) { printf("SearchStringBackward: curr = 0, line = %d\n", i); return -3;}
        if (FindString(curr->str, srchstr)) return i;
    }

    for (i = numline; i >= linenum; i--)
    {
        curr = GetLinePtr(i, &prev);
	if (!curr) { printf("SearchStringBackward: curr = 0, line = %d\n", i); return -3;}
        if (FindString(curr->str, srchstr)) return i;
    }

    return PARSING_ERROR;
}

// Parse the replace string command and extract rplcstr1 and rplcstr2
int ParseReplaceString(char *str)
{
    char *ptr;

    if (*str++ != 's') return 0;
    if (*str == 0) return 1;

    if (*str++ != '/') return 0;
    ptr = FindChar(str, '/');
    if (*ptr != '/') return 0;
    *ptr++ = 0;
    if (*str) strcpy(rplcstr1, str);
    else      strcpy(rplcstr1, srchstr);
    if (*ptr == 0) return 1;

    str = FindChar(ptr, '/');
    if (*str != '/') return 0;
    *str++ = 0;
    strcpy(rplcstr2, ptr);

    rflag = (*str == 'g');

    return 1;
}

// Store the results of the replace-string command on str1 into str2
int ReplaceString(char *str1, char *str2)
{
    int changeflag = 0;
    int len1 = strlen(rplcstr1);
    int len2 = strlen(rplcstr2);

    while (*str1)
    {
	if (strncmp(str1, rplcstr1, len1) == 0)
	{
	    changeflag = 1;
	    memcpy(str2, rplcstr2, len2);
	    str2 += len2;
	    str1 += len1;
	    if (!rflag) break;
	}
	else
	    *str2++ = *str1++;
    }
    while (*str1)
    {
        *str2++ = *str1++;
    }
    *str2 = 0;
    return changeflag;
}

// Extract a line number parameter from a command
char *GetLineNumber(char *str, int linenum, int *num)
{
    int val;

    if (str[0] == '.')
    {
	*num = linenum;
	str++;
    }
    else if (str[0] == '-')
    {
	str++;
	if (isdigit(*str))
	    val = getnum(&str);
	else
	    val = 1;
	*num = linenum - val;
    }
    else if (str[0] == '+')
    {
	str++;
	if (isdigit(*str))
	    val = getnum(&str);
	else
	    val = 1;
	*num = linenum + val;
    }
    else if (str[0] == '$')
    {
	*num = numline;
	str++;
    }
    else if (str[0] == '/')
    {
	*num = SearchStringForward(&str, linenum);
    }
    else if (str[0] == '?')
    {
        *num = SearchStringBackward(&str, linenum);
    }
    else if (isdigit(*str))
    {
	*num = getnum(&str);
    }
    else *num = UNSPECIFIED;
    if (*num != UNSPECIFIED &&
        (*num < 0 || *num > numline)) *num = PARSING_ERROR;
    return str;
}

// Extract the first and last numbers and the command from a command string
int ParseCommand(char **pstr, int linenum, int *first, int *last, int *cmd)
{
    char *str = *pstr;

    if (*str == ',')
    {
	*first = 1;
	*last = numline;
	*cmd = str[1];
        *pstr = ++str;
	return 1;
    }
    else if (*str == ';')
    {
	*first = linenum;
	*last = numline;
	*cmd = str[1];
        *pstr = ++str;
	return 1;
    }
    str = GetLineNumber(str, linenum, first);
    if (*first == PARSING_ERROR) return 0;
    if (*first == UNSPECIFIED)
    {
	if (numline == 0 && *str == 'i')
	{
	    *first = *last = 1;
	}
	else if (*str == 'g' || *str == 'v' || *str == 'w')
	{
	    *first = 1;
	    *last = numline;
	}
	else if (*str == 'j')
	{
	    *first = linenum;
	    *last = linenum + 1;
	    if (*last > numline) return 0;
	}
	else if (*str == 'r')
	{
	    *first = numline;
	    *last = numline;
	}
	else if (*str == 0 || *str == 'z')
	{
	    *first = *last = linenum + 1;
	    if (*first > numline) return 0;
	}
	else if (*str == '=')
	{
	    *first = *last = numline;
	}
	else
	{
	    *first = linenum;
	    *last = linenum;
	}
    }
    else if (*str == ',' || *str == ';')
    {
	str = GetLineNumber(str+1, linenum, last);
        if (*last == PARSING_ERROR) return 0;
	if (*last == UNSPECIFIED) *last = *first;
	if (*last < (*first)) return 0;
    }
    else *last = *first;
    *cmd = *str;
    *pstr = str;
    if (*first == 0)
    {
        str = FindChar("aq=EerxQhVfP", *cmd);
        if (*str == 0) return 0;
    }
    return 1;
}

// Print the help information
void help(void)
{
    printf("ED Commands\n");
    printf("\n");
    printf("Line Addresses\n");
    printf("n                - nth line\n");
    printf(".                - current line\n");
    printf("$                - last line\n");
    printf("-                - previous line\n");
    printf("-n               - nth previous line\n");
    printf("+                - next line\n");
    printf("+n               - nth next line\n");
    printf("n,m              - lines n through m\n");
    printf(",                - first through last lines\n");
    printf(";                - current through last lines\n");
    printf("/str/            - next line containing str\n");
    printf("?str?            - previous line containing str\n");
    printf("\n");
    printf("Editing Commands\n");
    printf("(.)a             - append text after the addressed line\n");
    printf("(.,.)c           - replace the addressed lines with text\n");
    printf("(.,.)d           - delete the addressed lines\n");
    printf("e file           - discard all lines and edit the specified file\n");
    printf("E file           - unconditionally discard all lines and edit specified file\n");
    printf("f file           - set the current file name and print the name\n");
    printf("(1,$)g/str/cmd   - execute a command on lines containing str\n");
    printf("h                - print help information\n");
    printf("(.)i             - insert text before the addressed line\n");
    printf("(.,.+1)j         - join the addressed lines\n");
    printf("(.,.)l           - print the addressed lines unambiguously\n");
    printf("(.,.)m(.)        - move the addressed lines to after the right-hand address\n");
    printf("(.,.)n           - print the addresses lines with line numbers\n");
    printf("(.,.)p           - print the addresses lines\n");
    printf("P                - toggle the prompt mode\n");
    printf("q                - quit\n");
    printf("($)r file        - read the specified file after the addressed line\n");
    printf("(.,.)s/st1/st2/g - substitute occurances of st1 with st2\n");
    printf("(.,.)t(.)        - copy the addressed lines to after the right-hand address\n");
    printf("(1,$)v/str/cmd   - execute a command on lines not containing str\n");
    printf("Vn               - set visualization mode to n lines or toggle on/off if no n\n");
    printf("(1,$)w file      - write the addressed lines to the specified file\n");
    printf("(.)x             - insert lines from the cut buffer after the addressed line\n");
    printf("(.,.)y           - yank the addressed lines to the cut buffer\n");
    printf("(.+1)zn          - scroll and print n lines from the addressed line\n");
    printf("($)=             - print the line number of the addressed line\n");
    printf("(.+1)            - print the addressed line\n");
}

// Print a visualization mode screen
void PrintScreen(int linenum)
{
    int i;
    int vlines1 = vlines - 1;
    int vlines2 = vlines/2;
    int first = linenum - vlines2;
    int last = first + vlines1;

    if (first < 1)
    {
	first = 1;
	last = first + vlines1;
        if (last > numline) last = numline;
    }
    else if (last > numline)
    {
	last = numline;
	first = last - vlines1;
	if (first < 1) first = 1;
    }

    printf("===============================================================\n");
    PrintLines(linenum, first, last, 'v');
    for (i = last - first; i < vlines1; i++) printf("~\n");
    printf("===============================================================\n");
}

// Empty a list and free the string links
void EmptyList(StringT *list)
{
    StringT *next;

    while (list)
    {
	next = list->next;
	sfree(list);
	list = next;
    }
}

// Delete lines from the text file list and move them to the yank list
int DeleteLines(int first, int last)
{
    int i;
    StringT *prev;
    StringT *curr = GetLinePtr(first, &prev);
    StringT *yank = 0;

    EmptyList(YankHead);
    YankHead = 0;

    for (i = first; i <= last; i++)
    {
        if (!curr) { printf("DeleteLines: line = %d\n", first); return 1; }
	if (!yank) YankHead = curr;
	else       yank->next = curr;
	yank = curr;
	curr = curr->next;
	numline--;
    }
    if (prev) prev->next = curr;
    else      ListHead   = curr;
    yank->next = 0;
    changed = 1;
    if (first > numline) first = numline;
    return first;
}

// Copy lines from the text file list to the yank list
int YankLines(int first, int last)
{
    int line;
    StringT *link;
    StringT *list = 0;
    StringT *curr = GetLinePtr(first, &link);

    EmptyList(YankHead);
    YankHead = 0;

    for (line = first; line <= last; line++)
    {
        if (!curr) { printf("YankLines: line = %d\n", line); return 1; }
	link = DuplicateString(curr);
	if (!link) break;
	if (!list) YankHead = link;
	else       list->next = link;
	curr = curr->next;
	list = link;
    }

    if (list) list->next = 0;
    return last;
}

// Copy the lines on the yank list to the text file list
int PushLines(int line)
{
    StringT *link;
    StringT *curr = YankHead;

    while (curr)
    {
	link = DuplicateString(curr);
	if (!link) break;
	InsertLine(link, ++line);
	curr = curr->next;
	changed = 1;
    }

    return line;
}

// Go into the text input mode and append after "linenum"
int AppendLines(int linenum)
{
    StringT *prev;
    StringT *curr = GetLinePtr(linenum+1, &prev);
    StringT *link;

    while (1)
    {
#ifdef CSPIN_FLAG
        gets(inbuf);
#else
        fgets(inbuf, 100, stdin);
#endif
        RemoveCRLF(inbuf);
        if (inbuf[0] == '.' && inbuf[1] == 0) break;
	link = AllocateString(inbuf);
	if (!link) break;
	link->next = curr;
	if (prev) prev->next = link;
	else      ListHead = link;
	prev = link;
	linenum++;
	numline++;
	changed = 1;
    }
    return linenum;
}

// Go into the text input mode and insert lines before "linenum"
int InsertLines(int linenum)
{
    int linenum1 = AppendLines(linenum - 1);
    if (linenum1 > linenum) linenum = linenum1;
    return linenum;
}

// Execute the replace-string command on multiple lines in the text buffer
int ReplaceStringCommand(char *inptr, int first, int last, int linenum)
{
    int i;
    int lastproc = linenum;
    int changeflag = 0;
    StringT *link;
    StringT *prev;
    StringT *curr = GetLinePtr(first, &prev);

    if (!ParseReplaceString(inptr))
    {
	printf("?\n");
	return last;
    }

    for (i = first; i <= last; i++)
    {
        if (!curr) { printf("ReplaceStringCommand: line = %d\n", i); return 1; }

        if (CheckMatch(curr->str, srchstr))
	{
	    prev = curr;
	    curr = curr->next;
	    continue;
	}

        if(ReplaceString(curr->str, inbuf))
	{
	    link = curr;
            curr = smalloc(strlen(inbuf) + 1);
	    if (!curr) return lastproc;
	    strcpy(curr->str, inbuf);
            if (prev) prev->next = curr;
            else      ListHead   = curr;
            curr->next = link->next;
	    sfree(link);
	    changeflag = 1;
	    changed = 1;
	}
	prev = curr;
	curr = curr->next;
        lastproc = i;
    }

    if (!changeflag) printf("?\n");

    return lastproc;
}

// Join lines together and move the original lines to the yank buffer
int JoinLines(int first, int last)
{
    int i;
    int len = 0;
    char *str;
    StringT *prev;
    StringT *curr = GetLinePtr(first, &prev);
    StringT *link = curr;
    StringT *yank = 0;

    EmptyList(YankHead);
    YankHead = 0;

    // Determine the total length of the selected lines
    for (i = first; i <= last; i++)
    {
        if (!link) { printf("JoinLines: line = %d\n", i); return 1; }
	len += strlen(link->str);
	link = link->next;
    }

    // Allocate a new string buffer and concatenate the strings
    link = smalloc(len + 1);
    if (!link) return first;
    str = link->str;
    *str = 0;
    for (i = first; i <= last; i++)
    {
        if (!curr) { printf("JoinLines: line = %d\n", i); return 1; }
	strcat(str, curr->str);
	if (!yank) YankHead = curr;
	else       yank->next = curr;
	yank = curr;
	curr = curr->next;
	numline--;
	changed = 1;
    }

    // Insert the new string
    if (prev) prev->next = link;
    else      ListHead   = link;
    link->next = curr;
    yank->next = 0;
    numline++;

    return first;
}

// Move or copy lines to after the destination address
int MoveLines(int first, int last, int dstaddr, int cmd)
{
    int i;
    StringT *prev;
    StringT *curr;
    StringT *link;

    int srcaddr, insertbefore;

    if (dstaddr < 0 || dstaddr > numline ||
        (first <= dstaddr && dstaddr < last))
    {
        printf("?\n");
    }
    else
    {
        srcaddr = first;
        insertbefore = dstaddr < first;
        for (i = first; i <= last; i++)
        {
            curr = GetLinePtr(srcaddr++, &prev);
            link = DuplicateString(curr);
            if (!link) break;
            InsertLine(link, ++dstaddr);
            if (insertbefore) srcaddr++;
            if (cmd == 'm')
            {
	        DeleteLine(--srcaddr);
	        if (!insertbefore) dstaddr--;
            }
        }
        changed = 1;
    }
    return dstaddr;
}

// Read a file after "linenum"
int ReadFile(char *fname, int linenum)
{
    int numchar = 0;
    FILE *infile = fopen(fname, "r");
    StringT *link;
    //char inbuf[100];

    if (infile)
    {
        while (fgets(inbuf, 200, infile))
        {
	    RemoveCRLF(inbuf);
	    link = AllocateString(inbuf);
	    if (!link) break;
	    InsertLine(link, ++linenum);
            numchar += strlen(inbuf);
        }
        fclose(infile);
    }

    printf("%d\n", numchar);

    return linenum;
}

// Main program entry
#ifdef CSPIN_FLAG
int main(void)
{
    int argc = 1;
    char **argv;
#else
int main(int argc, char **argv)
{
#endif
    FILE *outfile;
    int numchar = 0;
    int linenum = 0;
    int temp;
    int i;
    int first, last, cmd;
    char *LinePtr;
    char *inptr;
    char filename[100];
    StringT *curr;
    StringT *prev;

    // Initialization
    srchstr[0] = 0;
    rplcstr1[0] = 0;
    rplcstr2[0] = 0;
    bigbuf = malloc(MEMORY_SIZE);
    smallocinit(bigbuf, MEMORY_SIZE);
    ListHead = 0;
    YankHead = 0;

    // Copy the input file name if specified
    if (argc > 1) strcpy(filename, argv[1]);
    else filename[0] = 0;

    linenum = ReadFile(filename, 0);

    while (1)
    {
	// Print the screen and read a command
	gflag = 0;
	if (vmode) PrintScreen(linenum);
	CheckMemory();
	if (prompt) printf("*");
	gets(inbuf);
	inptr = inbuf;
	// Parse the command and extract the parameters
	if (!ParseCommand(&inptr, linenum, &first, &last, &cmd))
	{
	    printf("?\n");
	    continue;
	}
	// Parse a "g" or "v" command if present
	if (cmd == 'g' || cmd == 'v')
	{
	    inptr++;
	    if (GetString(&inptr))
	    {
		if (cmd == 'g')
		    gflag = 1;
		else
		    gflag = 2;
		cmd = *inptr;
		if (cmd == 0) cmd = 'p';
	    }
	    else
	    {
		printf("?\n");
		continue;
	    }
	}
	// Execute the command
        switch (cmd)
        {
	    // Print the current line if no command
            case 0:
	        if (numline == 0)
		{
		    printf("?\n");
		    continue;
		}
	        if (last >= 0) linenum = last;
	        else if (first >= 0) linenum = first;
	        curr = GetLinePtr(linenum, &prev);
	        LinePtr = curr->str;
                printf("%s\n", LinePtr);
                break;
	    // Quit unconditionally
            case 'Q':
                return 0;
                break;
	    // Check for changes and quit if OK
            case 'q':
	        if (changed == 0) return 0;
	        changed = 0;
	        printf("?\n");
                break;
	    // Reset the text buffer to the contents of the default file
#ifdef CSPIN_FLAG
            case 'e', 'E':
#else
	    case 'E':
	    case 'e':
#endif
	        if (changed && cmd == 'e')
	        {
	            changed = 0;
	            printf("?\n");
	        }
	        else
	        {
		    inptr = SkipChar(inptr + 1, ' ');
		    if (*inptr == 0 && *filename == 0)
		    {
		        printf("?\n");
		        continue;
		    }
		    if (*inptr) strcpy(filename, inptr);
                    EmptyList(YankHead);
                    YankHead = 0;
                    EmptyList(ListHead);
                    ListHead = 0;
                    numline = 0;
                    linenum = ReadFile(filename, 0);
		    changed = 0;
	        }
		break;
	    // Read the contents of a file after the specified line
	    case 'r':
	        inptr = SkipChar(inptr + 1, ' ');
	        if (*inptr == 0 && *filename == 0)
	        {
	            printf("?\n");
	            continue;
	        }
	        if (*inptr)
	        {
	            if (!filename[0]) strcpy(filename, inptr);
	        }
	        else inptr = filename;
		temp = numline;
	        linenum = ReadFile(inptr, first);
		if (numline != temp) changed = 1;
		break;
	    // Write the text buffer to a file
	    case 'w':
	        numchar = 0;
	        inptr = SkipChar(inptr + 1, ' ');
	        if (*inptr == 0 && *filename == 0)
	        {
	            printf("?\n");
	            continue;
	        }
	        if (*inptr)
	        {
	            if (!filename[0]) strcpy(filename, inptr);
	        }
	        else inptr = filename;
                outfile = fopen(inptr, "w");
	        if (!outfile)
	        {
	            printf("?\n");
	            continue;
	        }
	        for (i = first; i <= last; i++)
	        {
	            curr = GetLinePtr(i, &prev);
	            LinePtr = curr->str;
		    numchar += strlen(LinePtr);
	            fprintf(outfile, "%s\n", LinePtr);
	        }
                fclose(outfile);
	        printf("%d\n", numchar);
	        if (first == 1 && last == numline) changed = 0;
		break;
	    // Delete the specified lines
	    case 'd':
	        linenum = DeleteLines(first, last);
		break;
	    // Move or copy the specified lines
#ifdef CSPIN_FLAG
            case 'm', 't':
#else
	    case 'm':
	    case 't':
#endif
                if (inptr[1])
                    sscanf(inptr + 1, "%d", &temp);
                else
                    temp = linenum;
	        linenum = MoveLines(first, last, temp, cmd);
		break;
	    // Change the specified lines
	    case 'c':
	        temp = (last == numline);
	        linenum = DeleteLines(first, last);
		if (temp)
	            linenum = AppendLines(linenum);
		else
	            linenum = InsertLines(linenum);
		break;
	    // Append text after the specified line
	    case 'a':
	        linenum = AppendLines(first);
		break;
	    // Insert text before the specified line
	    case 'i':
	        linenum = InsertLines(first);
		break;
	    // Join together lines
	    case 'j':
	        linenum = JoinLines(first, last);
		break;
	    // Print lines in either the "p", "l" or "n" format
#ifdef CSPIN_FLAG
            case 'p', 'l', 'n':
#else
	    case 'p':
	    case 'l':
	    case 'n':
#endif
	        linenum = PrintLines(linenum, first, last, cmd);
		break;
	    // Print help information
	    case 'h':
	        help();
		break;
	    // Set the number of lines in the visualization mode or
	    // toggle the visualization mode if no lines are specified
	    case 'V':
                if (inptr[1])
                {
                    sscanf(inptr+1, "%d", &temp);
                    if (temp <= 0)
                    {
                        printf("?\n");
                        continue;
                    }
                    vlines = temp;
                    vmode = 1;
                }
                else
                    vmode = !vmode;
		break;
	    // Replace text in the specified lines
	    case 's':
	        linenum = ReplaceStringCommand(inptr, first, last, linenum);
		break;
	    // Push text from the yank buffer
	    case 'x':
	        linenum = PushLines(first);
		break;
	    // Yank lines to the yank buffer
	    case 'y':
	        linenum = YankLines(first, last);
		break;
	    // Set the number of lines to scroll if specified,
	    // and scroll forward
	    case 'z':
	        first = last;
	        if (inptr[1])
	        {
	            sscanf(inptr + 1, "%d", &temp);
		    if (temp <= 0)
		    {
		        printf("?\n");
		        continue;
		    }
		    scroll = temp;
	        }
	        last = first + scroll;
	        if (last > numline) last = numline;
	        PrintLines(linenum, first, last, 'p');
	        linenum = last;
		break;
	    // Set the default file name if specifed and print it
	    case 'f':
	        inptr = SkipChar(inptr + 1, ' ');
	        if (*inptr == 0 && *filename == 0)
	        {
	            printf("?\n");
	            continue;
	        }
	        if (*inptr) strcpy(filename, inptr);
	        printf("%s\n", filename);
		break;
	    // Print the number of lines, or the line number if specified
	    case '=':
	        printf("%d\n", last);
		break;
	    // Toggle the prompt mode
	    case 'P':
	        prompt = !prompt;
		break;
	    // Print "?" if not a valid command
	    default:
                printf("?\n");
		break;
	}
    }
    return 0;
}

#ifdef CSPIN_FLAG
#include "smalloc.c"
#endif
/*
+------------------------------------------------------------------------------------------------------------------------------+
|                                                   TERMS OF USE: MIT License                                                  |
+------------------------------------------------------------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    |
|files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    |
|modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software|
|is furnished to do so, subject to the following conditions:                                                                   |
|                                                                                                                              |
|The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.|
|                                                                                                                              |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          |
|WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         |
|COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   |
|ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         |
+------------------------------------------------------------------------------------------------------------------------------+
*/
